from .utils import read_init, logger_print, openers
from collections import defaultdict
import numpy as np
import dill




def select_motifs_single(project_name: str, filename: str):
    data = read_init(project_name)
    fmt = data.fmt
    names = data.motif_names
    # postfixes = set(data.motif_postfixes)
    del data
    
    with openers[fmt](f'{project_name}.fit.{fmt}', 'rb') as f:
        motif_score = dill.load(f).motif_mean
        std = np.linalg.pinv(motif_score.fim, hermitian=True).diagonal() ** 0.5
        motif_score = motif_score.mean / std ** 0.5
    
    
    
    try:
        with openers[fmt](f'{project_name}.predict.{fmt}', 'rb') as f:
            U = dill.load(f)
            filtered = U.filtered_motifs
            U = U.U.mean(axis=-1, keepdims=True)
        inds = np.arange(0, len(motif_score))
        inds = np.delete(inds, filtered)
        motif_score[inds] += U
    except FileNotFoundError:
        logger_print('Warning: no motif activities prediction found, using only means')
    motif_score = np.abs(motif_score).flatten()
    
    names_base = defaultdict(list)
    postfixes_d = dict()
    for i, name in enumerate(names):
        name = name.split('_')
        base = '_'.join(name[:-1])
        names_base[base].append(i)
        postfixes_d[i] = name[-1]
    names = list()
    for name, vals in names_base.items():
        i = max(vals, key=lambda x: motif_score[x])
        names.append(name + '_' + postfixes_d[i])
    with open(filename, 'w') as f:
        f.write('\n'.join(names))
        
        
        
    
    
    
    


# -*- coding: utf-8 -*-
from enum import Enum
from click import Context
from typer import Typer, Option, Argument
from typer.core import TyperGroup
from typing import List
from rich import print as rprint
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table
from pathlib import Path
from .fit import fit, ClusteringMode, calculate_fov, predict, GOFStat, TauMode, TauEstimation
from time import time
from .export import export_results


class OrderCommands(TyperGroup):
  def list_commands(self, ctx: Context):
    """Return list of commands in the order appear."""
    return list(self.commands)    # get commands using self.commands

app_old = Typer(rich_markup_mode='rich', cls=OrderCommands, add_completion=False)


@app_old.command('fit', help='Estimate variance parameters.')
def _fit(name: str = Argument(..., help='Project name.'),
         tau_mode: TauMode = Option(TauMode.mara, help='MARA or ISMARA model.'),
         tau_estimation: TauEstimation = Option(TauEstimation.fixed, help='Estimation method for tau'),
         tau_fix: float = Option(0.1, help='A value for tau to be fixed to given that [cyan]tau_estimation[/cyan] is [orange]fixed[/orange].'),
        clustering: ClusteringMode = Option(ClusteringMode.none, help='Clustering method.'),
        num_clusters: int = Option(200, help='Number of clusters if [orange]clustering[/orange] is not [orange]none[/orange].'),
        test_chromosomes: List[str] = Option(None, '--test-chromosomes', '-t', help='Test chromosomes'),
        gpu: bool = Option(False, help='Use GPU if available for most of computations.'), 
        gpu_decomposition: bool = Option(False, help='Use GPU if available or SVD decomposition.'), 
        x64: bool = Option(True, help='Use high precision algebra.')):
    """
    Fit a a mixture model parameters to data for the given project.
    """

    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Fitting model to the data...", total=None)
    p.start()
    fit(name, tau_mode=tau_mode, tau_estimation=tau_estimation, tau_fix=tau_fix, 
        clustering=clustering, num_clusters=num_clusters,
        gpu=gpu, test_chromosomes=test_chromosomes,
        gpu_decomposition=gpu_decomposition, x64=x64)
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')

@app_old.command('gof', help='Estimate GOFs given test/train data split. Provides test info only if [orange]test-chromosomes[/orange] is not None in [cyan]fit[/cyan].')
def _gof(name: str = Argument(..., help='Project name.'),
         # use_groups: bool = Option(False, help='Compute statistic for sammples aggragated across groups.'), 
         keep_motifs: Path = Option(None, help='Table with 2 columns: motif and status'),
         stat_type: GOFStat = Option(GOFStat.fov, help='Statistic type to compute'),
         gpu: bool = Option(False, help='Use GPU if available for most of computations.'), 
         x64: bool = Option(True, help='Use high precision algebra.')):
    """
    Fit a a mixture model parameters to data for the given project.
    """

    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Calculating FOVs...", total=None)
    p.start()
    res = calculate_fov(name, stat_type=stat_type, keep_motifs=keep_motifs, gpu=gpu, x64=x64)
    for name, res in res:
        if name:
            print(name)
        if stat_type == GOFStat.corr:
            title = 'Pearson correlation'
        else:
            title = 'Fraction of variance explained'
        if name:
            title = f'({name}) {title}'
        t = Table('Set', 'stat',
                  title=title)
        row = [f'{t.total:.6f}' for t in res.train]
        t.add_row('train', *row)
        if res.test is not None:
            row = [f'{t.total:.6f}' for t in res.test]
            t.add_row('test', *row)
        rprint(t)
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')

@app_old.command('predict', help='Estimate deviations of motif activities from their means.')
def _predict(name: str = Argument(..., help='Project name.'),
             gpu: bool = Option(False, help='Use GPU if available for most of computations.'), 
             x64: bool = Option(True, help='Use high precision algebra.')):
    """
    Fit a a mixture model parameters to data for the given project.
    """

    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Predicting motif activities...", total=None)
    p.start()
    predict(name, gpu=gpu, x64=x64)
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')


@app_old.command('export', help='Extract motif activities, parameter estimates.')
def _export(name: str = Argument(..., help='Project name.'),
            output_folder: Path = Argument(..., help='Output folder.')):
    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Exporting results...", total=None)
    p.start()
    export_results(name, output_folder)
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')



import jax.numpy as jnp
from jax import jit, grad
from jax.scipy.stats import norm
from jax.scipy.special import logsumexp, logit, expit
import pandas as pd
import numpy as np
from scipy.optimize import minimize
from functools import partial
from sklearn.mixture import GaussianMixture

def compute_leftmost_probability(Y):
    Y = Y.reshape(-1, 1)
    gmm = GaussianMixture(n_components=2, random_state=0)
    gmm.fit(Y)
    
    means = gmm.means_.flatten()
    leftmost_component_index = np.argmin(means)
    probas = gmm.predict_proba(Y)
    leftmost_probs = probas[:, leftmost_component_index]
    
    return leftmost_probs, gmm

def normax_logpdf(x: jnp.ndarray, mu: float, sigma: float, n: int):
    x = (x - mu) / sigma
    return jnp.log(n) - jnp.log(sigma) + norm.logpdf(x) + (n - 1) * norm.logcdf(x)

def logmixture(x: jnp.ndarray, mus: jnp.ndarray, sigmas: jnp.ndarray, w: float, n: int):
    logpdf1 = normax_logpdf(x, mus[0], sigmas[0], n)
    logpdf2 = normax_logpdf(x, mus[1], sigmas[1], n)
    w = jnp.array([w, 1 - w]).reshape(-1,1)
    logpdf = jnp.array([logpdf1, logpdf2])
    return logsumexp(logpdf, b=w, axis=0)


def transform(params, forward=True):
    mu = params[:2]
    sigma = params[2:4]
    w = params[-1:]
    if forward:
        sigma = sigma ** 2
        w = expit(w)
    else:
        sigma = sigma ** 0.5
        w = logit(w)
    return jnp.concatenate([mu, sigma, w])

def loglik(params: jnp.ndarray, x: jnp.ndarray, n: int):
    params = transform(params)
    mu = params[:2]
    sigma = params[2:4]
    w = params[-1]
    return -logmixture(x, mu, sigma, w, n).sum()

def filter_lowexp(expression: pd.DataFrame, cutoff=0.95, component_limit=0.6, max_mode=True,
                  fit_plot_filename=None, plot_dpi=200):
    expression = (expression - expression.mean()) / expression.std()
    if not max_mode:
        expression = expression.mean(axis=1).values
        probs, gmm = compute_leftmost_probability(expression)
        inds = probs < (1-cutoff)
        if fit_plot_filename:
            import matplotlib.pyplot as plt
            from matplotlib.collections import LineCollection
            import seaborn as sns
            x = np.array(sorted(expression))
            pdf = np.exp(gmm.score_samples(expression[:, None]))
            points = np.array([x, pdf]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            plt.figure(dpi=plot_dpi, )
            sns.histplot(expression, stat='density', color='grey')
            lc = LineCollection(segments, cmap='winter')
            lc.set_array(probs)
            lc.set_linewidth(3)
            line = plt.gca().add_collection(lc)
            plt.colorbar(line)
            plt.xlabel('Standardized expression')
            plt.tight_layout()
            plt.savefig(fit_plot_filename)
        return inds, probs
        
    expression_max = expression.max(axis=1).values

    mu = [-1.0, 0.0]
    sigmas = [1.0, 1.0]
    w = [0.5]
    x0 = jnp.array(mu + sigmas + w)
    x0 = transform(x0, False)
    fun = jit(partial(loglik, x=expression_max, n=expression.shape[1]))
    jac = jit(grad(fun))
    res = minimize(fun, x0, jac=jac)
    
    params = transform(res.x)
    mu = params[:2]
    sigma = params[2:4]
    w = params[-1]
    
    mode1 = minimize(lambda x: -normax_logpdf(x, mu[0], sigma[0], n=expression.shape[1]), x0=[0.0]).x
    mode2 = minimize(lambda x: -normax_logpdf(x, mu[1], sigma[1], n=expression.shape[1]), x0=[0.0]).x
    if mode1 > mode2:
        mu = mu[::-1]
        sigma = sigma[::-1]
        w = 1 - w
    
    inds = np.argsort(expression_max)
    inds_inv = np.empty_like(inds, dtype=int)
    inds_inv[inds] = np.arange(len(inds))
    x = expression_max[inds]
    logpdf1 = normax_logpdf(x, mu[0], sigma[0], n=expression.shape[1])
    logpdf2 = normax_logpdf(x, mu[1], sigma[1], n=expression.shape[1])
    pdf1 = jnp.exp(logpdf1)
    pdf2 = jnp.exp(logpdf2)
    ws = np.array(pdf1 / ((w * pdf1 + (1-w)*pdf2)) * w)
    
    if float(w) > component_limit:
        ws[:] = 1.0
    else:
        ws = 1 - ws
        if x[ws >= 0.5].mean() < x[ws < 0.5].mean():
            ws = 1 - ws
        j = np.argmax(ws)
        l = np.argmin(ws)
        ws[j:] = 1.0
        ws[:l] = 0.0
    
    k = 0
    for k in range(len(ws)):
        if ws[k] >= 1.0-cutoff:
            break
    if fit_plot_filename:
        import matplotlib.pyplot as plt
        from matplotlib.collections import LineCollection
        import seaborn as sns
        pdf = jnp.exp(logmixture(x, mu, sigma, w, n=expression.shape[1]))
        points = np.array([x, pdf]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)
        plt.figure(dpi=plot_dpi, )
        sns.histplot(expression_max, stat='density', color='grey')
        lc = LineCollection(segments, cmap='winter')
        lc.set_array(ws)
        lc.set_linewidth(3)
        line = plt.gca().add_collection(lc)
        plt.colorbar(line)
        plt.xlabel('Standardized expression')
        plt.tight_layout()
        plt.savefig(fit_plot_filename)
    ws = ws[inds_inv]
    inds = np.ones(len(expression), dtype=bool)
    inds[:k] = False
    # print(inds)
    # inds[:] = 1
    inds = inds[inds_inv]
    return inds, ws



# -*- coding: utf-8 -*-
from . import check_packages
from enum import Enum
from click import Context
from typer import Typer, Option, Argument
from typer.core import TyperGroup
from typing import List
from rich import print as rprint
from jax import __version__ as jax_version
from scipy import __version__ as scipy_version
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table
from .create import create_project
from pathlib import Path
from .fit import fit, ClusteringMode, calculate_fov, predict, GOFStat, GOFStatMode, GLSRefinement
from .grn import estimate_promoter_variance, grn
from .synthetic_data import generate_dataset
from time import time
from dill import __version__ as dill_version
from .export import export_results, export_loadings_product, Standardization, ANOVAType
from .export import export_contrast, export_posterior_anova
from . import __version__ as project_version
from .select import select_motifs_single
import json

# logging.getLogger("jax._src.xla_bridge").addFilter(logging.Filter("No GPU/TPU found, falling back to CPU."))
# logging.getLogger("jax._src.xla_bridge").addFilter(logging.Filter("An NVIDIA GPU may be present on this machine, but a CUDA-enabled jaxlib is not installed. Falling back to cpu."))

__all__ = ['main']

class Compression(str, Enum):
    lzma = 'lzma'
    gzip = 'gzip'
    bz2 = 'bz2'
    raw = 'raw'

class LoadingTransform(str, Enum):
    none = 'none'
    scale = 'scale'
    ecdf = 'ecdf'
    esf = 'esf'
    drist = 'drist'
    dristun = 'dristun'


class OrderCommands(TyperGroup):
  def list_commands(self, ctx: Context):
    """Return list of commands in the order appear."""
    return list(self.commands)    # get commands using self.commands

_DO_NOT_UPDATE_HISTORY = False

def update_history(name: str, command: str, **kwargs):
    if _DO_NOT_UPDATE_HISTORY:
        return
    try:
        with open(f'{name}.json', 'r') as f:
            d = json.load(f)
    except FileNotFoundError:
        d = dict()
    if command == 'create':
        d.clear()
        d['jax'] = jax_version
        d['maradoner'] = project_version
        d['scipy'] = scipy_version
        d['dill'] = dill_version
        d['name'] = name
    elif command == 'fit':
        for k in ('test', 'test_binom', 'difftest', 'combine', 'export', 'plot'):
            if k in d:
                del d[k]
        for k in list(d):
            if k.startswith('export'):
                del d[k]
    d[command] = kwargs
    with open(f'{name}.json', 'w') as f:
        json.dump(d, f, indent=4)
    
doc = f'''
[bold]MARADONER[/bold] version {project_version}: [bold]M[/bold]otif [bold]A[/bold]ctivity [bold]R[/bold]esponse [bold]A[/bold]nalaysis [bold]Done R[/bold]ight 
\b\n
\b\n
A typical [bold]MARADONER[/bold] session consists of sequential runs of [bold cyan]create[/bold cyan], [bold cyan]fit[/bold cyan],  [bold cyan]predict[/bold cyan] and, finally, \
[bold cyan]export[/bold cyan] commands. [bold]MARADONER[/bold] accepts files in the tabular format (.tsv or .csv, they also can come in gzipped-flavours), \
and requires the following inputs:
[bold orange]•[/bold orange] Promoter expression table of shape [blue]p[/blue] x [blue]s[/blue], where [blue]p[/blue] is a number of promoters and \
[blue]s[/blue] is a number of samples;
[bold orange]•[/bold orange] Matrix of loading coefficients of motifs onto promoters of shape [blue]p[/blue] x [blue]m[/blue], where [blue]m[/blue] \
is a number of motifs;
[bold orange]•[/bold orange] [i](Optional)[/i] Matrix of motif expression levels in log2 scale per sample of shape [blue]m[/blue] x [blue]s[/blue];
[bold orange]•[/bold orange] [i](Optional)[/i] JSON dictionary or a text file that collects samples into groups (if not supplied, it is assumed that \
each sample is a group of its own).
[red]Note:[/red] all tabular files must have named rows and columns.
All of the input files are supplied once at the [cyan]create[/cyan] stage. All of the commands are very customizable via numerous options, more \
details can be found in their respective helps, e.g.:
[magenta]>[/magenta] [cyan]maradoner fit --help[/cyan]

\b\n
If you found a bug or have any questions, feel free to contact us via
a) e-mail: [blue]iam@georgy.top[/blue] b) issue-tracker at [blue]github.com/autosome-ru/maradoner[/blue]
'''
app = Typer(rich_markup_mode='rich', cls=OrderCommands, add_completion=False, help=doc)

from .mara.main import app_old
app.add_typer(app_old, name='mara', help='Run classical (IS)MARA framework. For testing purposes.')

help_str = 'Initialize [bold]MARADONER[/bold] project initial files: do parsing and filtering of the input data.'

@app.command('create', help=help_str)
def _create(name: str = Argument(..., help='Project name. [bold]MARADONER[/bold] will produce files for internal usage that start with [cyan]'
                                            'name[/cyan].'),
            expression: Path = Argument(..., help='A path to the promoter expression table. Expression values are assumed to be in a log-scale.'),
            loading: List[Path] = Argument(..., help='A list (if applicable, separated by space) of filenames containing loading matrices. '),
            loading_transform: List[LoadingTransform] = Option([LoadingTransform.esf], '--loading-transform', '-t',
                                                               help='A type of transformation to apply to loading '
                                                                'matrices. [orange]ecdf[/orange] substitutes values in the table with empricical CDF,'
                                                                ' [orange]esf[/orange] with negative logarithm of the empirical survival function. '
                                                                '[orange]scale[/orange] just rescales all entries to [0, 1] range. '
                                                                '[orange]drist[/orange] will attempt to find an optimal monotonous transform of each column. '
                                                                '[orange]dristun[/orange] is [orange]drist[/orange] but the transform function is same for all columns/motifs.'),
            motif_expression: List[Path] = Option(None, help='A list of paths (of length equal to the number of loading matrices) of motif expression'
                                                  ' tables. All expression values are assumed to be in log2-scale.'),
            sample_groups: Path = Option(None, help='Either a JSON dictionary or a text file with a mapping between groups and sample names they'
                                          ' contain. If a text file, each line must start with a group name followed by space-separated sample names.'),
            filter_lowexp_w: float = Option(0.9, help='Truncation boundary for filtering out low-expressed promoters. The closer [orange]w[/orange]'
                                            ' to 1, the more promoters will be left in the dataset.'),
            filter_max_mode: bool = Option(True, help='Use max-mode of filtering. Max-mode keeps promoters that are active at least for some samples.'
                                                       ' If disabled, filtration using GMM on the averages will be ran instead.'),
            filter_plot: bool = Option(True, help='Expression plot with a fitted mixture that is used for filtering.'),
            loading_postfix: List[str] = Option(None, '--loading-postfix', '-p', 
                                                help='String postfixes will be appeneded to the motifs from each of the supplied loading matrices'),
            motif_filename: Path = Option(None, '--motif-filename', help='If provided, then only motifs with names present in this fill will be kept'),
            n_jobs: float = Option(0.5, help='Number of threads to use when reading datatables. If -1, maximal number of threads are used. If it is fractional in range (0, 1), '
                                   'the fraction maximal number is used.'),
            compression: Compression = Option(Compression.raw.value, help='Compression method used to store results.')):
    if type(compression) is Compression:
        compression = str(compression.value)
    if sample_groups:
        sample_groups = str(sample_groups)
    loading = list(map(str, loading))
    loading_transform = [x.value if issubclass(type(x), Enum) else str(x) for x in loading_transform]
    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Initializing project...", total=None)
    p.start()
    if filter_plot:
        filter_plot = f'{name}.filter-plot.png'
    r = create_project(name, expression, loading_matrix_filenames=loading, motif_expression_filenames=motif_expression, 
                       loading_matrix_transformations=loading_transform, sample_groups=sample_groups, 
                       promoter_filter_lowexp_cutoff=filter_lowexp_w,
                       promoter_filter_plot_filename=filter_plot,               
                       promoter_filter_max=filter_max_mode,
                       compression=compression, 
                       motif_postfixes=loading_postfix,
                       motif_names_filename=motif_filename,
                       n_jobs=n_jobs,
                       verbose=False)
    p.stop()
    dt = time() - t0
    p, s = r['expression'].shape
    m = r['loadings'].shape[1]
    g = len(r['groups'])
    rprint(f'Number of promoters: {p}, number of motifs: {m}, number of samples: {s}, number of groups: {g}')
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')
    

@app.command('fit', help='Estimate variance parameters and motif activities.')
def _fit(name: str = Argument(..., help='Project name.'),
          clustering: ClusteringMode = Option(ClusteringMode.none, help='Clustering method.'),
          num_clusters: int = Option(200, help='Number of clusters if [orange]clustering[/orange] is not [orange]none[/orange].'),
          test_promoters_filename: Path = Option(None, '--test-promoters-filename', help='Path to a text file containing gene/promoter names,'
                                                                                          ' separated by a newline character, to be used as a testing set.'),
          test_chromosomes: List[str] = Option(None, '--test-chromosomes', '-t', help='Test chromosome number, e.g. 3, X, etc.'),
          motif_variance: bool = Option(True,  help='Estimate individual motif variances or assume them all equal.'
                                                  'The latter makes method more similar to (IS)MARA approach and might be beneficial when the data is small.'),
          promoter_variance: bool = Option(False, help='Estiamte individual promoter variances or assume them all equal.'),
          refinement: GLSRefinement = Option(GLSRefinement.none, help='TODO help'),
          gpu: bool = Option(False, help='Use GPU if available for most of computations.'), 
          gpu_decomposition: bool = Option(False, help='Use GPU if available or SVD decomposition.'), 
          x64: bool = Option(True, help='Use high precision algebra.')):
    """
    Fit a a mixture model parameters to data for the given project.
    """

    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Fitting model to the data...", total=None)
    p.start()
    fit(name, clustering=clustering, num_clusters=num_clusters,
        gpu=gpu, test_chromosomes=test_chromosomes,
        motif_variance=motif_variance,
        promoter_variance=promoter_variance,
        refinement=refinement,
        test_promoters_filename=test_promoters_filename,
        gpu_decomposition=gpu_decomposition, x64=x64)
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')

@app.command('gof', help='Estimate GOFs given test/train data split. Provides test info only if [orange]test-chromosomes[/orange] is not None in [cyan]fit[/cyan].')
def _gof(name: str = Argument(..., help='Project name.'),
         use_groups: bool = Option(False, help='Compute statistic for samples aggragated across groups.'), 
         stat_type: GOFStat = Option(GOFStat.fov, help='Statistic type to compute'),
         stat_mode: GOFStatMode = Option(GOFStatMode.total, help='Whether to compute stats for residuals or accumulate their effects'),
         weights: bool = Option(True, help='Whether to use weighted statistics. Applicable only if promoter variance was estimated.'),
         gpu: bool = Option(False, help='Use GPU if available for most of computations.'), 
         x64: bool = Option(True, help='Use high precision algebra.')):
    """
    Fit a a mixture model parameters to data for the given project.
    """

    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Calculating FOVs...", total=None)
    p.start()
    res = calculate_fov(name, stat_mode=stat_mode, stat_type=stat_type, use_groups=use_groups, 
                        weights=weights,
                        gpu=gpu, x64=x64)
    if stat_type == GOFStat.corr:
        title = 'Pearson correlation'
    else:
        title = 'Fraction of variance explained'
    if stat_mode == GOFStatMode.residual:
        title += ' (Residual)'
    else:
        title += ' (Total)'
    t = Table('Set', 'centering', '+mean motif activity', '+ group-specific motif activity',
              title=title)
    row = [f'{t.total:.6f}' for t in res.train]
    
    t.add_row('train', *row)
    if res.test is not None:
        row = [f'{t.total:.6f}' for t in res.test]
        t.add_row('test', *row)
    p.stop()
    dt = time() - t0
    rprint(t)
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')

@app.command('predict', help='Estimate deviations of motif activities from their means.')
def _predict(name: str = Argument(..., help='Project name.'),
         filter_motifs: bool = Option(True, help='Do not predict deviations from motifs whose variance is low.'),
         filter_order: int = Option(7, help='Motif variance is considered low if it is [orange]filter-order[/orange] orders of magnitude smaller that a median motif variance.'),
         tau_search: bool = Option(False, help='Search for tau multiplier using CV'),
         cv_repeats: int = Option(3, help='CV repeats in [orange]RepeatedKFold[/orange]'),
         cv_splits: int = Option(5, help='CV splits in [orange]RepeatedKFold[/orange]'),
         tau_left: float = Option(0.1, help='Left bound for tau search'),
         tau_right: float = Option(1.0, help='Right bound for tau search. Values > 1 imply that we believe that variance could be underestimated'),
         tau_num: int = Option(10, help='Number of splits in np.linspace(tau_left, tau_right, tau_num)'),
         pinv: bool = Option(False, help='Use plain Moore-Penrose pseudoinverse with Tikhonov regularization for tau > 1. Applicable for testing purposes'),
         gpu: bool = Option(False, help='Use GPU if available for most of computations.'), 
         x64: bool = Option(True, help='Use high precision algebra.')):
    """
    Fit a a mixture model parameters to data for the given project.
    """

    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Predicting motif activities...", total=None)
    p.start()
    predict(name, tau_search=tau_search, cv_splits=cv_splits, cv_repeats=cv_repeats,
            tau_left=tau_left, tau_right=tau_right, tau_num=tau_num, pinv=pinv,
            filter_motifs=filter_motifs, filter_order=filter_order, gpu=gpu, x64=x64)
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')

@app.command('generate', help='Generate synthetic dataset for testing purporses.')
def _generate(output_folder: Path = Argument(..., help='Output folder.'),
                p: int = Option(2000, help='Number of promoters.'),
                m: int = Option(500, help='Number of motifs.'),
                g: int = Option(10, help='Number of groups.'),
                min_samples: int = Option(5, help='Minimal number of observations per each group.'),
                max_samples: int = Option(6, help='Maximal number of observations per each group.'),
                non_signficant_motifs_fraction: float = Option(0.20, help='Fraction of non-significant motifs.'),
                sigma_rel: float = Option(1e-1, help='Ratio of nu to sigma. The higher this value, the higher is FOV.'),
                means: bool = Option(True, help='Non-zero groupwise and promoter-wise intercepts'),
                seed: int = Option(1, help='Random seed.')
            ):
    t0 = time()
    pr = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    pr.add_task(description="Generating synthetic dataset...", total=None)
    pr.start()
    generate_dataset(folder=output_folder, p=p, m=m, g=g, min_samples=min_samples, max_samples=max_samples,
                     non_signficant_motifs_fraction=non_signficant_motifs_fraction, sigma_rel=sigma_rel,
                     means=means, 
                     B_cor=0, U_cor=0, E_cor=0, motif_cor=0,
                     seed=seed)
    pr.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')

@app.command('export', help='Extract motif activities, parameter estimates, FOVs and statistical tests.')
def _export(name: str = Argument(..., help='Project name.'),
            output_folder: Path = Argument(..., help='Output folder.'),
            std_mode: Standardization = Option(Standardization.full, help='Whether to standardize activities with plain variances or also decorrelate them.'),
            anova_mode: ANOVAType = Option(ANOVAType.positive, help='If negative, look for non-variative motifs'),
            weighted_zscore: bool = Option(False, help='Reciprocal variance weighted Z-scores'),
            alpha: float = Option(0.05, help='FDR alpha.'),
            log_counts: bool = Option(False, help='Export expression value predictions.'),
            log_counts_grouped: bool = Option(True, help='Whether to export expressions wrt to groups or samples.'),
            loadings_product: bool = Option(False, help='Export loading matrix-acitvity 3D tensor. This will produce num_of_groups tabular files.'),
            lp_hdf: bool = Option(True, help='Each loadings-product table will be stored in hdf format (occupies much less space than plain tsv) using float16 precision.'),
            lp_intercepts: bool = Option(True, help='Include motif means in the 3D tensor.'),
            lp_tsv_truncation: int = Option(4, help='Number of digits after a floating point to truncate. Decreases the output size of a tabular if [orange]lp-hdf[/orange] is disabled.')):
    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Exporting results...", total=None)
    p.start()
    export_results(name, output_folder, std_mode=std_mode, anova_mode=anova_mode, alpha=alpha,
                   weighted_zscore=weighted_zscore, export_counts=log_counts, counts_grouped=log_counts_grouped)
    p.stop()
    
    if loadings_product:
        p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
        p.add_task(description="Exporting results...", total=None)
        p.start()
        export_loadings_product(name, output_folder, use_hdf=lp_hdf, intercepts=lp_intercepts)
        p.stop()
    
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')

@app.command('anova', help='Run posterior ANOVA across groups of interest.')
def _anova(name: str = Argument(..., help='Project name.'),
            filename: Path = Argument(..., help='Output folder.'),
            groups: List[str] = Argument(..., help='Group names')):
    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Computing posterior ANOVA...", total=None)
    p.start()
    export_posterior_anova(name, filename, groups)
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')


__select_motif_doc = 'Selects best motif variants when the project was created from multiple loading matrices, each with an unique postfix.'\
                     'Best motif variant is select by comparing their standardized contribution/z-score of both intercepts and deviations.'\
                     'The resulting list of motif names with the appropriate postfixes will be then stored in [orange]filename[/orange].'\
                     'It''s assumed that [orange]filename[/orange] will be supplied as [orange]motif_filename[/orange] to the [bold cyan]create[/bold cyan].'
@app.command('select-motifs',
             help=__select_motif_doc)
def _select_motifs(name: str = Argument(..., help='Project name'),
                   filename: Path = Argument(..., help='Filename where a list of best motif variants will be stored')):
    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Selecting motifs...", total=None)
    p.start()
    select_motifs_single(name, filename)
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')


__grn_doc = '\bTests each promoter against each motif per each group. Some people call it a GRN.\n'\
            'It compares H_0 against H_1:\n'\
            ' [cyan]H_0: y = b u + b μ 1_s^T + e,[/cyan]\n'\
            " [cyan]H_1: y = (b u)' + (b μ)' 1_s^T + e,[/cyan]\n"\
            "where y is a vector of gene expression observations, apostraphe denotes the effect of the motif of interest being excluded from the model."\
            "If [purple]--no-means[/purple] is supplied, H_1 instead is\n"\
            " [cyan]H_1: y = (b u)' + [orange]b μ [/orange]1_s^T + e,[/cyan]\n "\
            "so we measure only effect that is due to the deviation u."
@app.command('grn',
             help=__grn_doc)
def _grn(name: str = Argument(..., help='Project name'),
         folder: Path = Argument(..., help='Output folder where results will be stored. In total, expect number_of_groups tables of size'
                                           ' comparable to the expression file size.'),
         hdf: bool = Option(True, help='Use HDF format instead of tar.gz files. Typically eats much less space'),
         stat: bool = Option(True, help='Save statistics alongside probabilities.'),
         prior_h1: float = Option(1/10, help='Prior belief on the expected fraction of motifs active per promoter.'),
         means: bool = Option(True, help='Include motif-specific means in H_1 model, otherwise only activities deviations are being tested.')):
    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Building GRN...", total=None)
    p.start()
    grn(name, output=folder, use_hdf=hdf, save_stat=stat, prior_h1=prior_h1, include_mean=means)
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')
    
__estimate_promvar_doc = 'Estimates each promoter variance for each group using Empirical Bayesian shrinkage.'\
                         ' The nature of the approach is similar to the one introduced in limma-voom. A very recommended step before computing GRN.'
@app.command('estimate-promoter-variance',
             help=__estimate_promvar_doc)
def _estimate_promoter_variance(name: str = Argument(..., help='Project name')):
    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Estimating each promoter's variance...", total=None)
    p.start()
    estimate_promoter_variance(name, )
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')
    
@app.command('contrast',
             help='Perform differential test given a constrasts vector.'
                 ' The contrast vector can be any algebraic string with variables being group names.'
                 ' For example, "[orange]healthy - ill[/orange]", "[orange](group_a + group_b + group_c) / 3 - group_d[/orange]".'
                 ' It computes a weighted algebraic sum of motif cctivities and then performs a series of Z-tests for each motif.')
def _contrast(name: str = Argument(..., help='Project name'),
                                contrast: str = Argument(...,
                                                          help='An algebraic expression defining the contrast vector.'),
                                output_folder: Path = Argument(..., help='Output folder.'),
                                postfix: str = Option(None, help='Postfix to add to the results file.')):
    t0 = time()
    p = Progress(SpinnerColumn(speed=0.5), TextColumn("[progress.description]{task.description}"), transient=True)
    p.add_task(description="Performing and saving a pairwise test...", total=None)
    p.start()
    export_contrast(name, output_folder, contrast, postfix)
    p.stop()
    dt = time() - t0
    rprint(f'[green][bold]✔️[/bold] Done![/green]\t time: {dt:.2f} s.')


def main():
    check_packages()
    app()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from pandas import DataFrame as DF
# add dot
from .utils import read_init, openers, ProjectData
from .fit import FOVResult, ActivitiesPrediction, FitResult, split_data, transform_data
from .grn import grn
from scipy.stats import norm, chi2, multivariate_normal, Covariance
from scipy.linalg import eigh, lapack, cholesky, solve
from statsmodels.stats import multitest
import numpy as np
from enum import Enum
from tqdm import tqdm
import multiprocessing as mp
from functools import partial
from scipy.integrate import quad
import math
import time
import dill
import os


class Standardization(str, Enum):
    full = 'full'
    std = 'std'
    
class ANOVAType(str, Enum):
    positive = 'positive'
    negative = 'negative'
    
def chol_inv(x: np.array):
    """
    Calculate invserse of matrix using Ctestholesky decomposition.

    Parameters
    ----------
    x : np.array
        Data with columns as variables and rows as observations.
    return_chol : bool
        Returns cholesky decomposition if True.

    Raises
    ------
    np.linalg.LinAlgError
        Rises when matrix is either ill-posed or not PD.

    Returns
    -------
    c : np.ndarray
        x^(-1).

    """
    c, info = lapack.dpotrf(x)
    if info:
        raise np.linalg.LinAlgError
    lapack.dpotri(c, overwrite_c=True)
    mx = c + c.T - np.diag(c.diagonal())
    return mx

    
class Information():
    eps = 1e-10
    
    def __init__(self, fim: np.ndarray, slc=None, use_preconditioner=False, filter_items=None):
        self.filter_items = filter_items
        if filter_items is not None:
            fim = np.delete(fim, filter_items, axis=0)
            fim = np.delete(fim, filter_items, axis=1)
        self.square_root_inv = self._square_root_inv(fim, slc, corr=True)
        precond = 1 / fim.diagonal() ** 0.5
        if not use_preconditioner:
            precond[:] = 1
        fim = precond.reshape(-1, 1) * fim * precond
        self.fim = fim
        self.precond = precond
        self.slice = slice(None, None) if slc is None else slc
    
    def _inv(self, x: np.ndarray):
        x = np.array(x)
        # t = np.linalg.eigh(x)
        try:
            x = chol_inv(x)
        except:
            print('Failed to compute inverse using Cholesky decomposition. ')
            print('This can be a sign of a numerical errors during parameters estimation.')
            print('Will use pseudo-inverse now. The minimal and maximal eigenvalues are:')
            # print(x.diagonal().min())
            assert np.allclose(x, x.T), x - x.T
            x = np.linalg.eigh(x)
            print(x[0].min(), x[0].max())
            # x = np.linalg.pinv(x, hermitian=True)
            x = x[1] * (1/np.clip(x[0], self.eps, float('inf'))) @ x[1].T
        return x
    
    def _square_root_inv(self, x: np.ndarray, slc=None, corr=True):
        x = self._inv(x)
        if corr:
            istd =  1 / x.diagonal() ** 0.5
            x = istd.reshape(-1, 1) * x * istd
        if slc is not None:
            x = x[slc, slc]
        try:
            x = cholesky(x)
        except:
            x = np.linalg.eigh(x)
            x = x[1] * x[0] ** (1 / 2) @ x[1].T
        return x
    
    def standardize(self, x: np.ndarray, 
                    mode: Standardization=Standardization.std,
                    return_std=True):
        if self.filter_items is not None:
            x = np.delete(x, self.filter_items)
        x = x / self.precond[self.slice]
        cov = self._inv(self.fim)
        cov = cov[self.slice, self.slice]
        std = cov.diagonal() ** 0.5
        if mode == mode.std:
            x /= std
        elif mode == mode.full:
            istd = 1 / std
            cor = istd.reshape(-1, 1) * cov * istd
            e = np.linalg.eigh(cor)
            T = istd.reshape(-1,1) * e[1] * e[0] ** (-0.5) @ e[1].T
            x = T @ x.reshape(-1, 1)
        if return_std:
            return x.flatten(), std * self.precond[self.slice]
        return x.flatten()
    
    def covariance(self):
        cov = self._inv(self.fim)
        cov = cov[self.slice, self.slice]
        return self.precond[self.slice].reshape(-1, 1) * cov * self.precond[self.slice]

    def correlation(self):
        cov = self.covariance()
        d = cov.diagonal() ** (-0.5)
        return d.reshape(-1, 1) * cov * d
    
    def cholesky_transform(self, x: np.ndarray):
        if self.square_root_inv is None:
            self.square_root_inv = self.square_root_inv(self.fim, self.slice, corr=True)
        return self.square_root_inv.T @ x
        



def export_fov(fovs: tuple[FOVResult], folder: str,
               promoter_names: list[str], sample_names: list[str]):
    os.makedirs(folder, exist_ok=True)
    cols = ['null', 'means', 'motif_means']
    fov_null, fov_means, fov_motif_means = fovs
    total = [fov_null.total, fov_means.total, fov_motif_means.total]
    DF(total, index=cols, columns=['FOV']).T.to_csv(os.path.join(folder, 'total.tsv'), sep='\t')
    promoters = [fov_null.promoter[:, None], fov_means.promoter[:, None], fov_motif_means.promoter[:, None]]
    promoters = np.concatenate(promoters, axis=-1)
    DF(promoters, index=promoter_names, columns=cols).to_csv(os.path.join(folder, 'promoters.tsv'), sep='\t')
    samples = [fov_null.sample[:, None], fov_means.sample[:, None], fov_motif_means.sample[:, None]]
    samples = np.concatenate(samples, axis=-1)
    DF(samples, index=sample_names, columns=cols).to_csv(os.path.join(folder, 'samples.tsv'), sep='\t')




def posterior_anova(activities: ActivitiesPrediction, fit: FitResult, 
                    B: np.ndarray, corr_stat=False, map_cov=False,
                    groups=None):
    precs = list()
    istds = list()
    covs = list()
    mean = 0.0
    bad_inds = np.zeros(activities.U.shape[0], dtype=bool)
    # for cov, U, nu in zip(activities.cov(), activities.U.T, fit.motif_variance.group):
    #     mot = fit.motif_variance.motif
    #     mot = np.delete(mot, activities.filtered_motifs)
    #     ind = mot * nu < cov.diagonal() + 1e-9
    #     bad_inds[ind] = True
    # mot = fit.motif_variance.motif
    # mot = np.delete(mot, activities.filtered_motifs)[~bad_inds]
    motif_variance = fit.motif_variance.motif
    if activities.filtered_motifs is not None:
        motif_variance = np.delete(motif_variance, activities.filtered_motifs)
        B = np.delete(B, activities.filtered_motifs, axis=1)
    U = activities.U
    if map_cov:
        # fit.motif_variance.m
        BTB = B.T @ B
        BTB_s = BTB * motif_variance ** 0.5
        BTB_s = BTB_s @ BTB_s.T
    for i, (cov, U, sigma, n, nu) in enumerate(zip(activities.cov(), U.T, 
                          activities._cov[-2], 
                          fit.error_variance.variance, fit.motif_variance.group)):
        # cov = cov[~bad_inds, ~bad_inds]
        # cov = cov[..., ~bad_inds]
        # cov = cov[~bad_inds]
        if groups and i not in groups:
            continue
        if map_cov:
            D = BTB_s * nu  + np.identity(len(BTB)) * sigma
            cov = cov @ D @ cov.T * n / sigma ** 2 
        covs.append(cov)
        # U = U[~bad_inds]
        # prec = np.linalg.inv(np.diag(mot * nu) - cov)
        prec = np.linalg.pinv(cov, hermitian=True)
        mean += prec @ U
        precs.append(prec)
    total_prec = sum(precs)
    total_cov = np.linalg.pinv(total_prec, hermitian=True)
    mean = total_cov @ mean
    stats = activities.U[~bad_inds] - mean.reshape(-1, 1)
    # if corr_stat:
    #     istd = 1 / total_cov.diagonal() ** 0.5
    #     total_cor = istd.reshape(-1, 1) * total_cov * istd
    #     stats = total_cor @ stats
    #     total_cov = total_cor @ total_cov @ total_cor
    # stats = (1 / total_cov.diagonal().reshape(-1, 1)) ** 0.5 * stats
    istds = [1 / c.diagonal() ** 0.5 for c in covs]
    istds = np.array(istds).T 
    if groups:
        stats = stats[:, np.array(groups)]
    stats = stats * istds
    stats = stats ** 2
    stats = stats.sum(axis=-1)
    pvalues = chi2.sf(stats, len(precs) - 1)
    fdr = multitest.multipletests(pvalues, alpha=0.05, method='fdr_by')[1] 
    return stats, pvalues, fdr, bad_inds
    

def calc_log_counts(data: ProjectData, fit: FitResult, activities: ActivitiesPrediction, groups: bool = True):
    sample_names = data.sample_names
    group_names = data.group_names
    promoter_names = data.promoter_names
    data = transform_data(data, helmert=False)
    B = data.B
    mu_m = B @ fit.motif_mean.mean.reshape(-1, 1)
    B = np.delete(B, activities.filtered_motifs, axis=1)
    mu_s = fit.sample_mean.mean.reshape(1, -1)
    mu_p = fit.promoter_mean.mean.reshape(-1, 1)
    if len(mu_p) != len(B):
        Y = data.Y  - mu_s - mu_m - B @ activities.U_raw
        D = (1 / fit.error_variance.variance)[data.group_inds_inv].reshape(-1, 1)
        mu_p = (Y @ D / (D.sum())).reshape(-1, 1)
    if not groups:
        cols = sample_names
        U = activities.U_raw
    else:
        lt = list()
        mu_s = mu_s.flatten()
        for inds in data.group_inds:
            lt.append(np.mean(mu_s[inds]))
        mu_s = np.array(lt).reshape(1, -1)
        cols = group_names
        U = activities.U
    log_counts = mu_s + mu_m + B @ U + mu_p
    return DF(log_counts, index=promoter_names, columns=cols)
    


def export_log_counts(output_folder: str, data: ProjectData, fit: FitResult,
                      activities: ActivitiesPrediction, group: bool = True):
    os.makedirs(output_folder, exist_ok=True)
    data_train, data_test = split_data(data, fit.promoter_inds_to_drop)
    log_counts = calc_log_counts(data_train, fit, activities, groups=group)
    log_counts.to_csv(os.path.join(output_folder, 'train.tsv'), sep='\t')
    if data_test:
        log_counts = calc_log_counts(data_test, fit, activities, groups=group)
        log_counts.to_csv(os.path.join(output_folder, 'test.tsv'), sep='\t')
    

def export_results(project_name: str, output_folder: str,
                   std_mode: Standardization, 
                   anova_mode: ANOVAType=ANOVAType.positive,
                   weighted_zscore=False,
                   alpha=0.05,
                   export_B: bool = False,
                   export_counts: bool = False,
                   counts_grouped: bool = True,
                   n_jobs=6):
    
    def calc_z_test(x):
        if anova_mode == ANOVAType.negative:
            import mpmath
            mpmath.mp.dps = 500
            pval = np.array([float(2 * mpmath.ncdf(t) - 1) for t in x])
        else:
            pval = 2 * norm.sf(np.abs(x))
        return pval

    data = read_init(project_name)
    fmt = data.fmt
    motif_names = data.motif_names
    prom_names = data.promoter_names
    if export_B:
        B = data.B
        B = DF(B, index=prom_names, columns=motif_names)
        os.makedirs(output_folder, exist_ok=True)
        B.to_csv(os.path.join(output_folder, 'B.tsv'), sep='\t')
    # del data
    with openers[fmt](f'{project_name}.fit.{fmt}', 'rb') as f:
        fit: FitResult = dill.load(f)
    if fit.promoter_inds_to_drop:
        prom_names = np.delete(prom_names, fit.promoter_inds_to_drop)
    group_names = fit.group_names
    with openers[fmt](f'{project_name}.predict.{fmt}', 'rb') as f:
        act: ActivitiesPrediction = dill.load(f)
    if act.filtered_motifs is not None:
        motif_names_filtered = np.delete(motif_names, act.filtered_motifs)
    else:
        motif_names_filtered = motif_names
    
    os.makedirs(output_folder, exist_ok=True)
    # grn(data, act, fit, os.path.join(output_folder, 'grn'))
    error_variance = fit.error_variance.variance
    # For backwards comptatability # TODO: rename promotor to promoter
    try:
        promoter_variance = fit.error_variance.promotor
    except:
        promoter_variance = None
    error_variance_fim = Information(fit.error_variance.fim)
    error_variance_stat, error_variance_std = error_variance_fim.standardize(error_variance, 
                                                                             mode=Standardization.std)
    
    motif_variance = fit.motif_variance.motif
    motif_variance_fim = Information(fit.motif_variance.fim, slice(None, len(motif_names_filtered)), 
                                     filter_items=act.filtered_motifs)
    motif_variance_stat, motif_variance_std = motif_variance_fim.standardize(motif_variance, 
                                                                             mode=Standardization.std)
    
    motif_group_variance = fit.motif_variance.group
    excluded_motif_group = fit.motif_variance.fixed_group
    motif_group_variance_fim = Information(fit.motif_variance.fim, slice(len(motif_names), None))
    motif_group_variance_std = motif_group_variance_fim.covariance().diagonal() ** 0.5
    
    
    motif_mean = fit.motif_mean.mean.flatten()
    motif_mean_fim = Information(fit.motif_mean.fim)
    motif_mean_stat, motif_mean_std = motif_mean_fim.standardize(motif_mean,
                                                                 mode=Standardization.std)
    
    promoter_mean = fit.promoter_mean.mean.flatten()
    # del fit
    
    if export_counts:
        folder = os.path.join(output_folder, 'log_counts')
        export_log_counts(folder, data, fit, act, counts_grouped)
    
    folder = os.path.join(output_folder, 'params')
    os.makedirs(folder, exist_ok=True)
    if os.path.isfile(f'{project_name}.promvar.{fmt}'):
        with openers[fmt](f'{project_name}.promvar.{fmt}', 'rb') as f:
            promvar: np.ndarray = dill.load(f)
        DF(promvar, index=prom_names, columns=group_names).to_csv(os.path.join(folder, 'promoter_variances.tsv'), sep='\t')
    if excluded_motif_group is not None:
        motif_group_variance_std = np.insert(motif_group_variance_std, excluded_motif_group, np.nan)
    DF(np.array([error_variance, error_variance_std, motif_group_variance, motif_group_variance_std]).T,
                index=group_names,
                columns=['sigma', 'sigma_std', 'nu', 'nu_std']).to_csv(os.path.join(folder, 'group_variances.tsv'),
                                                             sep='\t')
    s = 'motif\ttau\tstd\n' + '\n'.join(f'{a}\t{b}\t{c}' for a, b, c in zip(motif_names,
                                                                            motif_variance,
                                                                            motif_variance_std))
    with open(os.path.join(folder, 'motif_variances.tsv'), 'w') as f:
        f.write(s)
    
    if promoter_variance is None or promoter_variance.var() == 0:
        DF(promoter_mean, index=prom_names, columns=['mean']).to_csv(os.path.join(folder, 'promoter_means.tsv'),
                                                                     sep='\t')
    else:
        DF(np.array([promoter_mean, promoter_variance ** 0.5]).T, index=prom_names,
               columns=['mean', 'std']).to_csv(os.path.join(folder, 'promoter.tsv'),
                                               sep='\t')
    DF(np.array([motif_mean, motif_mean_std]).T,
       index=motif_names, columns=['mean', 'std']).to_csv(os.path.join(folder, 'motif_means.tsv'),
                                                          sep='\t')
    folder = os.path.join(folder, 'correlations')
    os.makedirs(folder, exist_ok=True)
    DF(fit.sample_mean.mean).to_csv(os.path.join(folder, 'sample_means.tsv'),
                                                                      sep='\t')
    DF(motif_mean_fim.correlation(), index=motif_names, columns=motif_names).to_csv(os.path.join(folder, 'motif_means.tsv'),
                                                                      sep='\t')
    DF(motif_variance_fim.correlation(), index=motif_names_filtered, columns=motif_names_filtered).to_csv(os.path.join(folder, 'motif_variances.tsv'),
                                                                      sep='\t')
    _group_names = group_names
    if excluded_motif_group is not None:
        _group_names = np.delete(_group_names, excluded_motif_group)
    DF(motif_group_variance_fim.correlation(), index=_group_names, columns=_group_names).to_csv(os.path.join(folder, 'motif_group_variances.tsv'),
                                                                      sep='\t')

    DF(error_variance_fim.correlation(), index=group_names, columns=group_names).to_csv(os.path.join(folder, 'error_variances.tsv'),
                                                                      sep='\t')
    
    
    folder = output_folder

    folder = os.path.join(output_folder, 'tests', 'prediction_based')
    os.makedirs(folder, exist_ok=True)

    stat, pvalue, fdr, bad_inds = posterior_anova(act, fit, B=data.B)
    motif_names_filtered = np.array(motif_names_filtered)[~bad_inds]
    anova = DF([stat, pvalue, fdr], columns=motif_names_filtered, index=['stat', 'p-value', 'FDR']).T
    anova.to_csv(os.path.join(folder, 'anova.tsv'), sep='\t')

    folder = os.path.join(output_folder, 'tests', 'asymptotics_based')
    os.makedirs(folder, exist_ok=True)
    
    anova_ass = motif_variance_stat
    pval = calc_z_test(anova_ass)

    fdrs = multitest.multipletests(pval, alpha=0.05, method='fdr_bh')[1]
    # lrt = 2 * fit.motif_variance.logratios
    # lrt_pvalues = chi2.sf(lrt, 1)
    # lrt_fdr = multitest.multipletests(lrt_pvalues, alpha=0.05, method='fdr_bh')[1]
    anova_ass = DF(np.array([anova_ass, pval, fdrs]).T, index=motif_names_filtered,
                   columns=['stat', 'p-value', 'FDR'])
    anova_ass.to_csv(os.path.join(folder, 'anova.tsv'), sep='\t')
    
    sign = motif_mean.flatten() / motif_mean_std
    neg = norm.cdf(sign)
    pos = norm.sf(sign)
    zero = chi2.cdf(sign ** 2, df=1)

    neg_fdr = multitest.multipletests(neg, alpha=0.05, method='fdr_bh')[1]
    pos_fdr = multitest.multipletests(pos, alpha=0.05, method='fdr_bh')[1]
    zero_fdr = multitest.multipletests(zero, alpha=0.05, method='fdr_bh')[1]
    sign_ass = DF(np.array([sign, zero, zero_fdr, neg, neg_fdr, pos, pos_fdr]).T, columns=['stat', 
                                                                                           'zero', 'fdr_zero',
                                                                                           'pvalue_neg', 'fdr_neg',
                                                                                           'pvalue_pos', 'fdr_pos'],
                  index=motif_names)
    sign_ass.to_csv(os.path.join(folder, 'sign.tsv'), sep='\t')
    
    folder = os.path.join(output_folder, 'activities')
    os.makedirs(folder, exist_ok=True)
    U = list()
    stds = list()
    for u, cov in zip(act.U.T, act.cov()):
        std = cov.diagonal() ** 0.5
        u = u / std
        U.append(u)
        stds.append(std)
    U = np.array(U).T
    DF(U, index=motif_names_filtered, columns=group_names).to_csv(os.path.join(folder, 'activity.tsv'), sep='\t')
    U = U ** 2
    if weighted_zscore:
        U_total = U.sum(axis=1, keepdims=True) / (1 / np.array(stds).T ** 2).sum(axis=1, keepdims=True)
    else:
        U_total = U.mean(axis=1, keepdims=True)
    
    U = np.hstack((U_total, U)) ** 0.5
    DF(U, index=motif_names_filtered,
       columns=['overall'] + list(group_names)).to_csv(os.path.join(folder, 'z_score.tsv'), sep='\t')
    DF(act.U_raw, index=motif_names_filtered, columns=data.sample_names).to_csv(os.path.join(folder, 'activity_raw.tsv'), sep='\t')
    
    
    
    if os.path.isfile(f'{project_name}.fov.{fmt}'):
        with open(f'{project_name}.fov.{fmt}', 'rb') as f:
            fov = dill.load(f)
            train = fov.train
            test = fov.test
        folder = os.path.join(output_folder, 'fov')
        if fov.grouped:
            sample_names = data.group_names
        else:
            sample_names = [None] * len(train[0].sample)
            for i, inds in enumerate(data.group_inds):
                name = data.group_names[i]
                for k, j in enumerate(inds):
                    sample_names[j] = f'{name}_{k+1}'
        if fit.promoter_inds_to_drop:
            promoter_names_train = np.delete(data.promoter_names, fit.promoter_inds_to_drop)
        else:
            promoter_names_train = data.promoter_names
        export_fov(train, os.path.join(folder, 'train'), promoter_names=promoter_names_train,
                   sample_names=sample_names)
        if test is not None:
            promoter_names_test = np.array(data.promoter_names)[fit.promoter_inds_to_drop]
            export_fov(test, os.path.join(folder, 'test'), promoter_names=promoter_names_test,
                       sample_names=sample_names)
            
def export_posterior_anova(project_name: str, filename: str,
                           groups: list[str]):
    

    data = read_init(project_name)
    fmt = data.fmt
    motif_names = data.motif_names
    prom_names = data.promoter_names
    with openers[fmt](f'{project_name}.fit.{fmt}', 'rb') as f:
        fit: FitResult = dill.load(f)
    if fit.promoter_inds_to_drop:
        prom_names = np.delete(prom_names, fit.promoter_inds_to_drop)
    group_names = fit.group_names
    with openers[fmt](f'{project_name}.predict.{fmt}', 'rb') as f:
        act: ActivitiesPrediction = dill.load(f)
    if act.filtered_motifs is not None:
        motif_names_filtered = np.delete(motif_names, act.filtered_motifs)
    else:
        motif_names_filtered = motif_names
    
    output_folder = os.path.split(filename)[0]
    if output_folder:
        os.makedirs(output_folder, exist_ok=True)
    groups = [i for i, n in enumerate(group_names) if n in groups]
    stat, pvalue, fdr, bad_inds = posterior_anova(act, fit, B=data.B, groups=groups)
    motif_names_filtered = np.array(motif_names_filtered)[~bad_inds]
    anova = DF([stat, pvalue, fdr], columns=motif_names_filtered, index=['stat', 'p-value', 'FDR']).T
    anova.to_csv(filename, sep='\t')

import ast
import sympy

def validate(node):
    """
    Validates that the AST only contains allowed operations: +, -, *, unary -, names, and numeric constants.
    """
    if isinstance(node, ast.Expression):
        return validate(node.body)
    if isinstance(node, ast.BinOp):
        validate(node.left)
        validate(node.right)
        allowed_ops = (ast.Add, ast.Sub, ast.Mult, ast.Div)
        if not isinstance(node.op, allowed_ops):
            raise ValueError(f"Unsupported operator: {type(node.op).__name__}")
    elif isinstance(node, ast.UnaryOp):
        validate(node.operand)
        if not isinstance(node.op, ast.USub):
            raise ValueError(f"Unsupported unary operator: {type(node.op).__name__}")
    elif isinstance(node, ast.Name):
        pass
    elif isinstance(node, ast.Constant):
        if not isinstance(node.value, (int, float)):
            raise ValueError("Non-numeric constant")
    else:
        raise ValueError(f"Unsupported node type: {type(node).__name__}")



def parse_linear_expression(expr_str):
    """
    Parses a linear algebraic expression string with only +, -, *, parentheses, and constants.
    Raises ValueError for unsupported operations (e.g., ^, /, sin) or non-linear terms (e.g., a*b, a^2).
    Returns a dictionary of variable coefficients.
    
    Example:
    >>> parse_linear_expression("(my_variable_1 * 2 - my_variable_2) - (3 * a - b)")
    {'a': -3, 'b': 1, 'my_variable_1': 2, 'my_variable_2': -1}
    
    >>> parse_linear_expression("(my_variable_1 - my_variable_2) - (a - b) * 2")
    {'a': -2, 'b': 2, 'my_variable_1': 1, 'my_variable_2': -1}
    """
    preprocessed = expr_str.replace(' ', '').replace('^', '**')
    try:
        tree = ast.parse(preprocessed, mode='eval')
    except SyntaxError as e:
        raise ValueError(f"Invalid expression syntax: {e}")
    
    validate(tree)
    
    expr = sympy.sympify(preprocessed)
    expr = sympy.expand(expr)
    symbols = sorted(expr.free_symbols, key=str)
    coeffs = {str(s): int(expr.coeff(s)) if expr.coeff(s).is_integer else float(expr.coeff(s)) 
              for s in symbols}
    
    # Ensure it's linear
    if symbols:
        poly = expr.as_poly(symbols)
        if poly.total_degree() > 1:
            raise ValueError("Expression must be linear (no products of variables or powers >1)")
    
    return dict(sorted(coeffs.items(), key=lambda x: x[0]))


def export_contrast(project_name: str,  output_folder: str,
                    contrasts: str, filename_postfix: str = None):
    if filename_postfix is None:
        postfix = str()
    else:
        postfix = '_' + filename_postfix
    data = read_init(project_name)
    fmt = data.fmt
    motif_names = data.motif_names
    prom_names = data.promoter_names
    del data
    with openers[fmt](f'{project_name}.fit.{fmt}', 'rb') as f:
        fit: FitResult = dill.load(f)
    if fit.promoter_inds_to_drop:
        prom_names = np.delete(prom_names, fit.promoter_inds_to_drop)
    group_names = fit.group_names
    with openers[fmt](f'{project_name}.predict.{fmt}', 'rb') as f:
        act: ActivitiesPrediction = dill.load(f)
    if act.filtered_motifs is not None:
        motif_names_filtered = np.delete(motif_names, act.filtered_motifs)
    else:
        motif_names_filtered = motif_names
    contrasts = parse_linear_expression(contrasts)
    variances = list()
    for k, cov in enumerate(act.cov()):
        variances.append(cov.diagonal())
    variances = np.array(variances) 
    coeffs = np.zeros_like(variances)
    inds = np.zeros(len(variances), dtype=bool)
    for c, v in contrasts.items():
        i = group_names.index(c)
        coeffs[i] = v
        inds[i] = True
    U = act.U.T
    U = U[inds]
    variances = variances[inds]
    coeffs = coeffs[inds]
    variances = variances * coeffs ** 2
    z_stat = (U * coeffs).sum(axis=0) / variances.sum(axis=0) ** 0.5
    # z_stat =  (U_b - U_a) / (variances_a + variances_b) ** 0.5
    pval = 2 * norm.sf(np.abs(z_stat))
    fdr = multitest.multipletests(pval, alpha=0.05, method='fdr_by')[1]
    data = np.array([z_stat, pval, fdr]).T
    filename = os.path.join(output_folder, f'constrast{postfix}.tsv')
    os.makedirs(output_folder, exist_ok=True)
    DF(data, columns=['z_stat', 'pval', 'fdr'],
       index=motif_names_filtered).to_csv(filename, sep='\t')

         
def export_loadings_product(project_name: str, output_folder: str,
                            use_hdf: bool = True, intercepts: bool = True,
                            tsv_truncation=4):
    

    data = read_init(project_name)
    fmt = data.fmt
    motif_names = data.motif_names
    prom_names = data.promoter_names
    # del data
    with openers[fmt](f'{project_name}.fit.{fmt}', 'rb') as f:
        fit: FitResult = dill.load(f)
    if fit.promoter_inds_to_drop:
        prom_names = np.delete(prom_names, fit.promoter_inds_to_drop)
    group_names = fit.group_names
    with openers[fmt](f'{project_name}.predict.{fmt}', 'rb') as f:
        act: ActivitiesPrediction = dill.load(f)
    
    output_folder = os.path.join(output_folder, 'loadings-product')
    os.makedirs(output_folder, exist_ok=True)
    
    U = act.U
    B = data.B
    mu = fit.motif_mean.mean 
    
    if act.filtered_motifs is not None:
        motif_names = np.delete(motif_names, act.filtered_motifs)
        B = np.delete(B, act.filtered_motifs, axis=1)
        mu = np.delete(mu, act.filtered_motifs)
    BM = B * mu
    for name, U in zip(group_names, U.T):
        effect = B * U
        if intercepts:
            effect += BM
        if use_hdf:
            effect = effect.astype(np.half)
            filename = os.path.join(output_folder, f'{name}.hdf')
            DF(data=effect, index=prom_names, columns=motif_names).to_hdf(filename, key='lrt', mode='w', complevel=4)
        else:
            filename = os.path.join(output_folder, f'{name}.tsv')
            DF(data=effect, index=prom_names, columns=motif_names).to_csv(filename, sep='\t',
                                                                          float_format=f'%.{tsv_truncation}f')  
        
        
    


import numpy as np
import jax.numpy as jnp
import jax
from dataclasses import dataclass
from functools import partial
from sklearn.model_selection import RepeatedKFold
from collections import defaultdict
from scipy.optimize import minimize_scalar
from enum import Enum
from ..utils import read_init, ProjectData, logger_print, openers
from ..fit import LowrankDecomposition, lowrank_decomposition, TransformedData, \
    ClusteringMode, cluster_data, FOVResult, TestResult
import dill


class GOFStat(str, Enum):
    fov = 'fov'
    corr = 'corr'

class GOFStatMode(str, Enum):
    residual = 'residual'
    total = 'total'
    
class TauEstimation(str, Enum):
    fixed = 'fixed'
    mle = 'mle'

class TauMode(str, Enum):
    mara = 'mara'
    ismara = 'ismara'



@dataclass(frozen=True)
class ErrorVarianceEstimates:
    variance: np.ndarray

@dataclass(frozen=True)
class MotifVarianceEstimates:
    variance: np.ndarray


@dataclass(frozen=True)
class FitResult:
    error_variance: ErrorVarianceEstimates
    motif_variance: MotifVarianceEstimates
    B_decomposition: LowrankDecomposition
    group_names: list
    clustering: np.ndarray = None
    clustered_B: np.ndarray = None
    promoter_inds_to_drop: list = None


def transform_data(data, std_y=False, std_b=False) -> TransformedData:
    Y = data.Y - (data.Y.mean(axis=0, keepdims=True) + data.Y.mean(axis=1, keepdims=True) - data.Y.mean())
    B = data.B - data.B.mean(axis=0, keepdims=True)
    group_inds_inv = list()
    group_inds = data.group_inds
    d = dict()
    for i, items in enumerate(group_inds):
        for j in items:
            d[j] = i
    for i in sorted(d.keys()):
        group_inds_inv.append(d[i])
    group_inds_inv = np.array(group_inds_inv)
    return TransformedData(Y=Y, B=B, 
                           group_inds=group_inds,
                           group_inds_inv=group_inds_inv)



def estimate_error_variance(data: TransformedData,
                            B_decomposition: LowrankDecomposition) -> ErrorVarianceEstimates:
    # Y = B_decomposition.null_Q.T @ data.Y
    Y = B_decomposition.null_space_transform(data.Y)
    variance = (Y ** 2).mean(axis=0)
    return ErrorVarianceEstimates(variance)


def calc_tau(tau: float, error_variance: np.ndarray, mode: TauMode):
    if mode == mode.mara:
        taus = tau * np.ones_like(error_variance)
    else:
        taus = tau / error_variance
    return taus

def loglik_tau(tau: float, Sigma: np.ndarray, Y_hat: np.ndarray,
               error_variance: np.ndarray, mode: TauMode) -> float:
    vec = 0
    logdet = 0
    taus = calc_tau(tau, error_variance, mode)
    for sigma, tau, y in zip(error_variance, taus, Y_hat.T):
        S = tau / sigma * Sigma + 1
        vec += (y ** 2 / S).sum() * (tau / sigma ** 2)
        logdet += S.sum()
    return -vec + logdet
        
def estimate_motif_variance(data: TransformedData, B_decomposition: LowrankDecomposition,
                            error_variance: ErrorVarianceEstimates,
                            estimation_method: TauEstimation,
                            mode: TauMode,
                            fixed_value=0.1) -> MotifVarianceEstimates:
    if estimation_method == estimation_method.fixed:
        tau = calc_tau(fixed_value, error_variance.variance, mode)
        return MotifVarianceEstimates(tau)
    Sigma = B_decomposition.S ** 2
    Q = B_decomposition.V.T
    Y_hat = Q.T @ data.B.T @ data.Y
    fun = partial(loglik_tau, Sigma=Sigma, Y_hat=Y_hat, error_variance=error_variance.variance,
                  mode=mode)
    tau = calc_tau(minimize_scalar(fun, bounds=(0.0, error_variance.variance.max() * 10)).x, error_variance.variance, mode)
    return MotifVarianceEstimates(tau)
    

@dataclass(frozen=True)
class ActivitiesPrediction:
    U: np.ndarray
    variance: np.ndarray
    clustering: tuple[np.ndarray, np.ndarray] = None
            

def predict_activities(data: TransformedData, fit: FitResult, 
                       gpu=False, verbose=True) -> ActivitiesPrediction:
    U = list()
    variance = list()
    
    B_decomposition = fit.B_decomposition
    if gpu:
        device = jax.devices()
    else:
        device = jax.devices('cpu')
    device = next(iter(device))
    with jax.default_device(device):
        Y_hat = data.B.T @ data.Y
        Sigma = B_decomposition.S ** 2
        Q = B_decomposition.V.T
        for y, sigma, tau in zip(Y_hat.T, fit.error_variance.variance, fit.motif_variance.variance):
            Z = Q / (1 / tau + Sigma / sigma) @ Q.T
            u = Z @ y / sigma 
            variance.append(Z.diagonal())
            U.append(u)
    U = np.array(U).T
    variance = np.array(variance).T
    return ActivitiesPrediction(U, variance)


def fit(project: str, tau_mode: TauMode, tau_estimation: TauEstimation,
        tau_fix: float, clustering: ClusteringMode,
        num_clusters: int, test_chromosomes: list,
        gpu: bool, gpu_decomposition: bool, x64=True,
        verbose=True, dump=True) -> ActivitiesPrediction:
    if x64:
        jax.config.update("jax_enable_x64", True)
    data = read_init(project)
    fmt = data.fmt
    group_names = data.group_names
    if clustering != clustering.none:
        logger_print('Clustering data...', verbose)
    data.B, clustering = cluster_data(data.B, mode=clustering, 
                                      num_clusters=num_clusters)
    if test_chromosomes:
        import re
        pattern = re.compile(r'chr([0-9XYM]+|\d+)')
        
        test_chromosomes = set(test_chromosomes)
        promoter_inds_to_drop = [i for i, p in enumerate(data.promoter_names) 
                                 if pattern.search(p).group() in test_chromosomes]
        data.Y = np.delete(data.Y, promoter_inds_to_drop, axis=0)
        data.B = np.delete(data.B, promoter_inds_to_drop, axis=0)
    else:
        promoter_inds_to_drop = None
    logger_print('Transforming data...', verbose)
    data = transform_data(data)
    if gpu_decomposition:
        device = jax.devices()
    else:
        device = jax.devices('cpu')
    device = next(iter(device))

    logger_print('Computing low-rank decompositions of the loading matrix...', verbose)
    with jax.default_device(device):
        B_decomposition = lowrank_decomposition(data.B)
    if gpu:
        device = jax.devices()
    else:
        device = jax.devices('cpu')
    device = next(iter(device))
    # print(data.B.shape, data_orig.B.shape)
    with jax.default_device(device):

        logger_print('Estimating error variances...', verbose)
        error_variance = estimate_error_variance(data, B_decomposition)       
        
        logger_print('Estimating variances of motif activities...', verbose)
        motif_variance = estimate_motif_variance(data, B_decomposition,
                                                  error_variance=error_variance,
                                                  mode=tau_mode, estimation_method=tau_estimation,
                                                  fixed_value=tau_fix)
        
    
    res = FitResult(error_variance=error_variance, motif_variance=motif_variance,
                    clustering=clustering, B_decomposition=B_decomposition,
                    group_names=group_names, promoter_inds_to_drop=promoter_inds_to_drop)    
    if dump:
        with openers[fmt](f'{project}.old.fit.{fmt}', 'wb') as f:
            dill.dump(res, f)
    return res

def split_data(data: ProjectData, inds: list) -> tuple[ProjectData, ProjectData]:
    if not inds:
        return data, None
    B = data.B
    Y = data.Y
    # Y = Y -  (Y.mean(axis=0, keepdims=True) + Y.mean(axis=1, keepdims=True) - Y.mean())
    # B = B - B.mean(axis=0, keepdims=True)
    B_d = np.delete(B, inds, axis=0)
    B = B[inds]
    Y_d = np.delete(Y, inds, axis=0)
    Y = Y[inds]
    promoter_names_d = np.delete(data.promoter_names, inds)
    promoter_names = list(np.array(data.promoter_names)[inds])
    data_d = ProjectData(Y=Y_d, B=B_d, K=data.K, weights=data.weights,
                         group_inds=data.group_inds, group_names=data.group_names,
                         motif_names=data.motif_names, promoter_names=promoter_names_d,
                         motif_postfixes=data.motif_postfixes, sample_names=data.sample_names,
                         fmt=data.fmt)
    data = ProjectData(Y=Y, B=B, K=data.K, weights=data.weights,
                         group_inds=data.group_inds, group_names=data.group_names,
                         motif_names=data.motif_names, promoter_names=promoter_names,
                         motif_postfixes=data.motif_postfixes, sample_names=data.sample_names,
                         fmt=data.fmt)
    return data_d, data

def predict(project: str,  gpu: bool, x64=True, dump=True):
    if x64:
        jax.config.update("jax_enable_x64", True)
    data = read_init(project)
    fmt = data.fmt
    with openers[fmt](f'{project}.old.fit.{fmt}', 'rb') as f:
        fit = dill.load(f)
    data, _ = split_data(data, fit.promoter_inds_to_drop)
    data = transform_data(data)
    if gpu:
        device = jax.devices()
    else:
        device = jax.devices('cpu')
    device = next(iter(device))
    with jax.default_device(device):
        activities = predict_activities(data, fit)
    if dump:
        with openers[fmt](f'{project}.old.predict.{fmt}', 'wb') as f:
            dill.dump(activities, f)
    return activities

def _cor(a, b, axis=1):
    a_centered = a - a.mean(axis=axis, keepdims=True)
    b_centered = b - b.mean(axis=axis, keepdims=True)
    numerator = np.sum(a_centered * b_centered, axis=axis)
    denominator = np.sqrt(np.sum(a_centered**2, axis=axis) * np.sum(b_centered**2, axis=axis))
    return numerator / denominator

def calculate_fov(project: str, gpu: bool, 
                  stat_type: GOFStat, keep_motifs: str, x64=True,
                  verbose=True, dump=True):
    keep_motifs = None
    def calc_fov(data: TransformedData, fit: FitResult,
                 activities: ActivitiesPrediction, keep_motifs=None, Bs=None) -> tuple[FOVResult]:
        def sub(Y, effects) -> FOVResult:
            if stat_type == stat_type.fov:
                Y1 = Y - effects
                Y = Y - Y.mean()
                Y1 = Y1 - Y1.mean()
                Y = Y ** 2
                Y1 = Y1 ** 2
                prom = 1 - Y1.mean(axis=1) / Y.mean(axis=1)
                sample = 1 - Y1.mean(axis=0) / Y.mean(axis=0)
                total = 1 - Y1.mean() / Y.mean()
            elif stat_type == stat_type.corr:
                # Y = Y - Y.mean(axis=0, keepdims=True) - Y.mean(axis=1, keepdims=True) + Y.mean()
                total = np.corrcoef(Y.flatten(), effects.flatten())[0, 1]
                prom = _cor(Y, effects, axis=1)
                sample = _cor(Y, effects, axis=0)
            return FOVResult(total, prom, sample)
        m2 = Bs[1].mean(axis=0, keepdims=True)
        m0 = Bs[1].mean()
        
        # Bs = None
        if Bs is None:
            # data = transform_data(data)
            B = data.B if activities.clustering is None else activities.clustering[0]
            Y = data.Y
            U = activities.U
        else:
            B_ = Bs[0]
            B = data.B
            Y = data.Y
            U = np.linalg.pinv(Bs[0]) @ (Bs[1] - Bs[1].mean(axis=0, keepdims=True) - Bs[1].mean(axis=1, keepdims=True) + Bs[1].mean())
            # B = np.hstack((B, np.ones((len(B), 1))))
            # U = np.linalg.pinv(np.hstack((Bs[0], np.ones((len(Bs[0]), 1))))) @ Bs[1]
            # m0 = Y.mean(axis=1, keepdims=True)
            # m2 = 0
            # m0 = 0
        if keep_motifs is not None:
            B = B[:, keep_motifs]
            U = U[keep_motifs]
        # B = B - B.mean(axis=0, keepdims=True)
        d = B @ U
        # m2 = Y.mean(axis=0, keepdims=True)
        # m0 = Y.mean()
        m = (Y.mean(axis=1, keepdims=True) + m2 - m0)
        d = d + m
        stat_0 = sub(Y, d)
        return stat_0,
    data = read_init(project)
    fmt = data.fmt
    motif_names = data.motif_names
    if keep_motifs:
        import datatable as dt
        df = dt.fread(keep_motifs).to_pandas().groupby('status')
        keep_motifs = list()
        for name, motifs in df:
            inds = list()
            for mot in motifs.iloc[:, 0]:
                try:
                    i = motif_names.index(mot)
                    inds.append(i)
                except ValueError:
                    print(f'Motif {mot} not found in the project.')
            keep_motifs.append((name, np.array(inds, dtype=int)))
    else:
        keep_motifs = [(None, None)]
    with openers[fmt](f'{project}.old.fit.{fmt}', 'rb') as f:
        fit = dill.load(f)
    with openers[fmt](f'{project}.old.predict.{fmt}', 'rb') as f:
        activities = dill.load(f)
    data, data_test = split_data(data, fit.promoter_inds_to_drop)
    if x64:
        jax.config.update("jax_enable_x64", True)
    # data = transform_data(data, helmert=False)
    # if data_test is not None:
    #     data_test = transform_data(data_test, helmert=False)
    if gpu:
        device = jax.devices()
    else:
        device = jax.devices('cpu')
    device = next(iter(device))
    results = list()
    for status_name, motifs in keep_motifs:
        if status_name:
            status_name = f'{status_name} ({len(motifs)})'
        with jax.default_device(device):
    
            if data_test is not None:
                test_FOV = calc_fov(data=data_test, fit=fit, activities=activities, keep_motifs=motifs,
                                    Bs=(data.B, data.Y)
                                    )
            train_FOV = calc_fov(data=data, fit=fit, activities=activities, keep_motifs=motifs,
                                 Bs=(data.B, data.Y)
                                 )
        if data_test is None:
            test_FOV = None
        res = TestResult(train_FOV, test_FOV, grouped=False)
        results.append((status_name, res))
    with openers[fmt](f'{project}.old.fov.{fmt}', 'wb') as f:
        dill.dump(results, f)
    return results
        
        
        


from time import time
import matplotlib.pyplot as plt
from scipy.optimize import minimize_scalar
import pandas as pd
from tqdm import tqdm
import scipy.linalg.lapack as lapack
import numpy as np
import scipy.stats as st
from collections import defaultdict
from sklearn.model_selection import RepeatedKFold
from koptimizer import KOptimizer
from jax import config
from os import environ
import jax.numpy as jnp


from jax.numpy.linalg import eigh, svd



def chol_inv(x: np.array):
    """
    Calculate invserse of matrix using Cholesky decomposition.

    Parameters
    ----------
    x : np.array
        Data with columns as variables and rows as observations.

    Raises
    ------
    np.linalg.LinAlgError
        Rises when matrix is either ill-posed or not PD.

    Returns
    -------
    c : np.ndarray
        x^(-1).

    """
    c, info = lapack.dpotrf(x)
    if info:
        raise np.linalg.LinAlgError
    lapack.dpotri(c, overwrite_c=1)
    inds = np.tri(len(c), k=-1, dtype=bool)
    c[inds] = c.T[inds]
    return c


def lowrank_decomposition(X: np.ndarray, rel_eps=1e-10):
    q, s, v = [np.array(t) for t in svd(X)]
    max_sv = max(s)
    n = len(s)
    for r in range(n):
        if s[r] / max_sv < rel_eps:
            r -= 1
            break
    r += 1
    s = s[:r] ** 0.5
    null_q = q[:, r:]
    null_v = v[r:]
    q = q[:, :r]
    v = v[:r]
    return (q, s ** 2, v), (null_q, null_v)


def _nullspace_eigh(X):
    if type(X) not in (tuple, list):
        X = jnp.array(X)
        P0 = jnp.identity(X.shape[0], dtype=float) - X @ jnp.linalg.inv(X.T @ X) @ X.T
    else:
        Q = X[0]
        Q = jnp.array(Q)
        I = jnp.identity(Q.shape[0], dtype=float)
        P0 = I - Q @ Q.T
    return np.array(eigh(P0))


def estimate_groupwise_variance(Y: list[np.ndarray], B: np.ndarray, groups: list[tuple]) -> np.ndarray:
    """
    Estimate the diagonal variance matrix using the Restricted Maximum Likelihood method.

    Parameters
    ----------
    Y : List[np.ndarray]
        List of size g (number of groups), where each item is an array of shape (p, n), where p is a number of promoters and n is a number of samples.
    B : np.ndarray or tuple[np.ndarray, np.ndarray, np.ndarray]
        Loadings matrix or an SVD decomposition as returned by lowrank decomposition.

    Returns
    -------
    var_reml : np.ndarray
        Array of estimated variances.
    """
    if type(B) in (jnp.ndarray, np.ndarray):
        eigs, q = _nullspace_eigh(B)
        for j in range(len(eigs)):
            if eigs[j] > 0.8:
                break
        P = q[:, j:].T
    else:
        P = B[0].T
    P = jnp.array(P)
    Y = jnp.array(Y)
    var_reml = jnp.array([jnp.mean((P @ Y.at[..., inds].get()) ** 2) for inds in groups])
    return var_reml


def _estimate_transform_matrices(B_svd: tuple[np.ndarray, np.ndarray, np.ndarray], U_mults: np.ndarray):
    res = list()
    Q, D, V = B_svd
    dvt = jnp.array(D.reshape(-1, 1) * V)
    for mult in tqdm(U_mults):
        mult = jnp.array(mult)
        mx = dvt * mult ** 2 @ dvt.T
        qdv, _ = lowrank_decomposition(mx, rel_eps=1e-12)
        res.append((qdv[0].T, qdv[1]))
    return res


def estimate_sigma(Y: list[np.ndarray], B_svd: tuple[np.ndarray, np.ndarray, np.ndarray], variances: np.ndarray, U_mults: np.ndarray,
                   return_transformed=True, per_group=True, groups=None) -> np.ndarray:
    if not per_group and groups is not None:
        n = sum(map(len, groups))
        var = np.empty(n, dtype=float)
        for inds, v in zip(groups, variances):
            var[inds] = v
        variances = var
    if len(U_mults) < Y.shape[1]:
        n = sum(map(len, groups))
        mults = np.empty((n, U_mults.shape[-1]), dtype=float)
        for inds, mult in zip(groups, U_mults):
            mults[inds] = mult
        U_mults = mults
    P, eigs, _ = B_svd
    Yt = Y.T @ P
    trs = _estimate_transform_matrices(B_svd, U_mults)
    Yt = [P @ Y[..., np.newaxis] for Y, (P, _) in zip(Yt, trs)]
    eigs = [eigs for _, eigs in trs]
    aux = Yt
    bounds = (0.0, variances.max() * 10)

    def loglik(Y: list[jnp.ndarray], sigma: float, S: jnp.ndarray, eigs: list[jnp.ndarray]):
        loglik = 0
        for Y, sigma_g, eigs in zip(Y, S, eigs):
            R = sigma * eigs + sigma_g
            Y = (R.reshape(-1, 1)) ** (-0.5) * Y
            loglik += -jnp.einsum('ij,ij->', Y, Y)
            loglik += -Y.shape[1] * jnp.log(R).sum()
        return loglik

    if not per_group:
        def f(sigma): return -loglik(Yt, sigma, variances, eigs)
        r = minimize_scalar(f)
        return (r.x, aux) if return_transformed else r.x

    if groups is None:
        res = np.empty(Y.shape[1])
        for i in tqdm(list(range(Y.shape[1]))):
            def f(sigma): return -loglik(Yt[i:i + 1], sigma, variances[i:i + 1], eigs[i:i+1])
            r = minimize_scalar(f, bounds=bounds)
            res[i] = r.x
    else:
        res = np.empty(len(groups))
        for i in tqdm(list(range(len(res)))):
            inds = groups[i]
            def f(sigma): return -loglik([jnp.array(Yt[j]) for j in inds], sigma, jnp.array(variances[i:i + 1].repeat(len(inds))), [eigs[j] for j in inds])
            r = minimize_scalar(f, bounds=bounds)
            res[i] = r.x

    return (res, aux) if return_transformed else res


def estimate_u_map(Y, B_svd, variances, sigmas: np.ndarray, U_mults: np.ndarray, tau=1, groups=None,
                   k_fold=4, n_rep=1, B_orig=None):
    if groups is not None:
        Yt = np.empty((len(groups), len(Y)), dtype=float)
        U_mults_t = np.empty((len(groups), U_mults.shape[-1]), dtype=float)
        n = np.empty((len(groups),), dtype=float)
        for i, inds in enumerate(groups):
            n[i] = len(inds)
            Yt[i] = Y[:, inds].sum(axis=-1)
            U_mults_t[i] = U_mults[inds].mean(axis=0)
        U_mults = U_mults_t
        Y = Yt
    if len(U_mults) > Y.shape[1]:
        n = sum(map(len, groups))
        mults = np.empty((n, U_mults.shape[-1]), dtype=float)
        for inds, mult in zip(groups, U_mults):
            mults[inds] = mult
        U_mults = mults
    Q, D, V = [jnp.array(t) for t in B_svd]
    Vt = V.T * D
    BB = Vt @ Vt.T
    inds = jnp.diag_indices_from(BB)

    def _est(Y, Q, BB, tau: float) -> np.ndarray:
        nonlocal n
        Y = (Y @ Q) @ Vt.T
        U = list()
        U_std = list()
        stds = list()
        stds2 = list()
        for Y, sigma_g, sigma, mult, nt in zip(Y, variances, sigmas, U_mults, n):
            ratio = jnp.exp(jnp.log(sigma_g) - jnp.log(sigma) - jnp.log(tau))
            cov = nt * BB 
            cov = cov.at[inds].add(ratio * mult ** -2)
            cov = cov / sigma_g
            cov = jnp.linalg.pinv(cov, hermitian=True)
            U.append(cov @ Y[..., np.newaxis] / sigma_g )
            stds.append(cov.diagonal() ** 0.5)
            s = stds[-1]
            cor = (1/s) * cov * (1/s).reshape(-1,1)
            d, q = jnp.linalg.eigh(cor)
            p = q * (1/(d ** 0.5)) @ q.T
            W = p * (1/s).reshape(1,-1)

            U_std.append(W@U[-1])
            cov2 = cov @ ((BB * mult.reshape(1,-1) ** 2 @ BB) / sigma + BB / sigma_g) @ cov 
            stds2.append(cov2.diagonal() ** 0.5)
        return np.concatenate(U, axis=1), np.concatenate(U_std, axis=1), np.array(stds).T, np.array(stds2).T
    return _est(Y, Q, BB, tau)


def preprocess_data(Y, B, U_mult, std_y=False, std_b=False, b_cutoff=1e-9, weights=None, _means_est=False):
    if std_y:
        Y = Y / Y.std(axis=0, keepdims=True)
    n_samples = Y.shape[1]
    if _means_est:
        Ym = Y - Y.mean(axis=0, keepdims=True)
    if n_samples < 3:
        Y = Y - Y.mean(axis=0, keepdims=True)
    else:
        Y = Y - Y.mean(axis=0, keepdims=True) - Y.mean(axis=1, keepdims=True) + Y.mean()
    B = B.copy()
    B[B < b_cutoff] = 0.0
    min_b, max_b = B.min(axis=0, keepdims=True), B.max(axis=0, keepdims=True)
    inds = ((max_b - min_b) > 1e-4).flatten()
    inds[:] = True
    B = B[:, inds]
    min_b = min_b[:, inds]
    max_b = max_b[:, inds]
    U_mult = U_mult[:, inds]
    # B -= B.mean(axis=0, keepdims=True)
    std1 = B.std(axis=1, keepdims=True)
    if std_b:
        B /= std1
    if weights is not None:
        weights = weights.reshape(-1, 1) ** -0.5
        Y = weights * Y
        B = weights * B
    if _means_est:
        Y = (Y, Ym)
    return Y, B, U_mult, inds



